/*
 *  tslib/src/ts_read_raw_module.c
 *
 *  Original version:
 *  Copyright (C) 2001 Russell King.
 *
 *  Rewritten for the Linux input device API:
 *  Copyright (C) 2002 Nicolas Pitre
 *
 * This file is placed under the LGPL.  Please see the file
 * COPYING for more details.
 *
 * $Id: input-raw.c,v 1.5 2005/02/26 01:47:23 kergoth Exp $
 *
 * Read raw pressure, x, y, and timestamp from a touchscreen device.
 */
#include "config.h"

#include <errno.h>
#include <stdio.h>

#include <stdlib.h>
//#ifdef HAVE_UNISTD_H
//#include <unistd.h>
//#endif
#include <sys/time.h>
#include <sys/types.h>
#include <linux/input.h>
#include <termios.h>
#include <ctype.h>
#include <unistd.h>


#ifndef EV_SYN /* 2.4 kernel headers */
# define EV_SYN 0x00
#endif

#include "tslib-private.h"

#define PACKET_LEN 9

//#define DEBUG

struct tslib_input {
	struct tslib_module_info module;

	int	current_x;
	int	current_y;
	int	current_p;
	int misc;

	int	sane_fd;
	int	using_syn;
};


int WacomFlush(int fd)
{
	char ch[16];
	fd_set fdsRead;
	struct timeval timeout;
#ifdef DEBUG
	printf("flush serial port.\n");
#endif
	
	if (tcflush(fd, TCIFLUSH) == 0)
		return 0;

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;

	while (1)
	{
		FD_ZERO(&fdsRead);
		FD_SET(fd, &fdsRead);
		if (select(FD_SETSIZE,&fdsRead,NULL,NULL,&timeout) <= 0)
			break;
		read(fd,&ch,sizeof(ch));
	}

	return 0;
}

static int SerialReadRaw(int fd, unsigned char* puchData)
{
	int nXfer;
	unsigned char* pPos, *pEnd;
	unsigned int uCnt;
	
#ifdef DEBUG
	printf("serial read raw.\n");	
#endif 

	for (uCnt=0; uCnt<PACKET_LEN; uCnt+=nXfer)
	{
		nXfer = read(fd,puchData+uCnt,PACKET_LEN-uCnt);
		if (nXfer <= 0)
		{
#ifdef DEBUG
			fprintf(stderr,"read serial port failed: %s\n", strerror(errno));
#endif 
			return nXfer;
		}

		/* look for high-bit */
		pEnd = puchData + nXfer;
		for (pPos=puchData; pPos<pEnd; ++pPos)
		{
			if (*pPos & 0x80) break;
		}

		/* not where it was expected? fix it. */
		if (pPos != puchData) 
		{
#ifdef DEBUG
			fprintf(stderr,"Bad high bit, discarding %d bytes\n",pPos - puchData);
#endif

			/* copy remaining bytes down, if any */
			memmove(puchData,pPos,(nXfer - (pPos-puchData)));
			nXfer -= pPos - puchData;
		}
	}

	return (signed)uCnt;
}



static int SerialParseTabletPC(const unsigned char* puchData, unsigned int uLength, struct ts_sample *samp)
{
	int   prox = 0, eraser = 0, button = 0; //x=0, y=0, ;

	/* Tablet PC Supports: 256 pressure, eraser, 1/2 side-switch */
#ifdef DEBUG
	printf("parsing tablet pc.\n");
#endif 

	if (uLength != 9 && uLength != 11) {return -1; }
	
	prox = puchData[0] & 0x20 ? 1 : 0;
	if (prox)
	{
		samp->pressure = ((puchData[6] & 0x01) << 7) | (puchData[5] & 0x7F);

		samp->x = (((int)puchData[6] & 0x60) >> 5) |
					((int)puchData[2] << 2) |
					((int)puchData[1] << 9);
		samp->y = (((int)puchData[6] & 0x18) >> 3) |
					((int)puchData[4] << 2) |
					((int)puchData[3] << 9);
					
		eraser = (puchData[0] & 0x04) ? 1 : 0;
		
		samp->tool = eraser ? WACOM_TOOL_ERASER : WACOM_TOOL_PEN;
		
		button |= (puchData[0] & 0x01) ? LeftButton : 0;

		/* pen has right button, eraser has none */
		if(samp->tool == WACOM_TOOL_PEN)
			button |= (puchData[0] & 0x02) ? RightButton : 0;
				
		samp->button = button;
		
		//gettimeofday(&(samp->tv), NULL);
			
		return 1;
	}

	return 0;
}

static int ts_input_read_tty(struct tslib_module_info *inf,
			 struct ts_sample *samp, int nr)
{
	struct tslib_input *i = (struct tslib_input *)inf;
	struct tsdev *ts = inf->dev;
	int ret = nr;
	int total = 0;
	unsigned char puchData[PACKET_LEN];
	int iSize = 0;
	
	if(i->sane_fd == 0)
	{
		WacomFlush(ts->fd);
		i->sane_fd = 1;
	}
	
	while(total < nr)
	{
		if( PACKET_LEN > (iSize = SerialReadRaw(ts->fd, puchData)))
		{
#ifdef DEBUG
			fprintf(stderr, "raw data size is wrong: %d\n", iSize);
#endif 
			return  0;
		}
			
		ret = SerialParseTabletPC(puchData, PACKET_LEN, samp);
		if(ret > 0)
		{
			total += ret;
			samp  += ret;
		}
		else 
			return ret;
	}

	return ret = total;
}



static int check_fd(struct tslib_input *i)
{
	struct tsdev *ts = i->module.dev;
	int version;
	u_int32_t bit;
	u_int64_t absbit;

	if (! ((ioctl(ts->fd, EVIOCGVERSION, &version) >= 0) &&
		(version == EV_VERSION) &&
		(ioctl(ts->fd, EVIOCGBIT(0, sizeof(bit) * 8), &bit) >= 0) &&
		(bit & (1 << EV_ABS)) &&
		(ioctl(ts->fd, EVIOCGBIT(EV_ABS, sizeof(absbit) * 8), &absbit) >= 0) &&
		(absbit & (1 << ABS_X)) &&
		(absbit & (1 << ABS_Y)) && (absbit & (1 << ABS_PRESSURE)))) {
		fprintf(stderr, "selected device is not a touchscreen I understand\n");
		return -1;
	}

	if (bit & (1 << EV_SYN))
		i->using_syn = 1;

	return 0;
}

static int ts_input_read(struct tslib_module_info *inf,
			 struct ts_sample *samp, int nr)
{
	struct tslib_input *i = (struct tslib_input *)inf;
	struct tsdev *ts = inf->dev;
	struct input_event ev;
	int ret = nr;
	int total = 0;
	static int button = 0;
	static int tool = 0;
	
	if (i->sane_fd == 0)
		i->sane_fd = check_fd(i);

	if (i->sane_fd == -1)
		return 0;

	if (i->using_syn) {
		while (total < nr) {
			ret = read(ts->fd, &ev, sizeof(struct input_event));
			if (ret < (int)sizeof(struct input_event)) {
				total = -1;
				break;
			}
	
			switch (ev.type) {
			case EV_KEY:
				switch (ev.code) {

				case BTN_TOOL_RUBBER:
					tool  = WACOM_TOOL_ERASER;
		#ifdef DEBUG
					fprintf(stdout, "The tool is eraser, value = %d\n", ev.value);
		#endif		 /*DEBUG*/
					break;

				case BTN_TOOL_PEN:
					tool  = WACOM_TOOL_PEN;
		#ifdef DEBUG
					fprintf(stdout, "The tool is pen, value = %d\n", ev.value);
		#endif		 /*DEBUG*/
					break;

				case BTN_TOUCH:
					button = ev.value ? LeftButton : 0;
		#ifdef DEBUG
					fprintf(stdout, "left button, value = %d\n", ev.value);
		#endif		 /*DEBUG*/
					break;

				case BTN_STYLUS:
					button = ev.value ? RightButton : 0;
		#ifdef DEBUG
					fprintf(stdout, "right button, value = %d\n", ev.value);
		#endif		 /*DEBUG*/
					break;

				case BTN_STYLUS2:
		#ifdef DEBUG
					fprintf(stdout, "middle button, value = %d\n", ev.value);
		#endif		/*DEBUG*/
					break;
				}
				break;
			case EV_SYN:
				/* Fill out a new complete event */
				samp->x = i->current_x;
				samp->y = i->current_y;
				samp->pressure = i->current_p;
				samp->tool = tool;
				samp->button  = button;
				samp->tv = ev.time;
	#ifdef DEBUG
				fprintf(stdout, "RAW---------------------> %d %d %d %d\n",
						misc, samp->x, samp->y, samp->pressure);
	#endif		 /*DEBUG*/
				if(i->misc)
				{
					samp++;
					total++;
				}
				else
					return ret = total;
					
				break;
			case EV_ABS:
				switch (ev.code) {
				case ABS_Y:
					i->current_x = ev.value;
					break;
				case ABS_X:
					i->current_y = ev.value;
					break;
				case ABS_PRESSURE:
					i->current_p = ev.value;
					break;
				case ABS_MISC:
					i->misc = ev.value;
					break;
				}
				break;
			}
		}
		ret = total;
	} else {
		unsigned char *p = (unsigned char *) &ev;
		int len = sizeof(struct input_event);
	
		while (total < nr) {
			ret = read(ts->fd, p, len);
			if (ret == -1) {
				if (errno == EINTR) {
					continue;
				}
				break;
			}
	
			if (ret < (int)sizeof(struct input_event)) {
				/* short read
				 * restart read to get the rest of the event
				 */
				p += ret;
				len -= ret;
				continue;
			}
			/* successful read of a whole event */
	
			if (ev.type == EV_ABS) {
				switch (ev.code) {
				case ABS_X:
					if (ev.value != 0) {
						samp->x = i->current_x = ev.value;
						samp->y = i->current_y;
						samp->pressure = i->current_p;
					} else {
						fprintf(stderr, "tslib: dropped x = 0\n");
						continue;
					}
					break;
				case ABS_Y:
					if (ev.value != 0) {
						samp->x = i->current_x;
						samp->y = i->current_y = ev.value;
						samp->pressure = i->current_p;
					} else {
						fprintf(stderr, "tslib: dropped y = 0\n");
						continue;
					}
					break;
				case ABS_PRESSURE:
					samp->x = i->current_x;
					samp->y = i->current_y;
					samp->pressure = i->current_p = ev.value;
					break;
				}
				samp->tv = ev.time;
	#ifdef DEBUG
				fprintf(stderr, "RAW---------------------------> %d %d %d\n",
					samp->x, samp->y, samp->pressure);
	#endif	 /*DEBUG*/
				samp++;
				total++;
			} else if (ev.type == EV_KEY) {
				switch (ev.code) {
				case BTN_TOUCH:
					if (ev.value == 0) {
						/* pen up */
						samp->x = 0;
						samp->y = 0;
						samp->pressure = 0;
						samp->tv = ev.time;
						samp++;
						total++;
					}
					break;
				}
			} else {
				fprintf(stderr, "tslib: Unknown event type %d\n", ev.type);
			}
			p = (unsigned char *) &ev;
		}
		ret = total;
	}

	return ret;
}

static int ts_input_fini(struct tslib_module_info *inf)
{
	free(inf);
	return 0;
}

static const struct tslib_ops __ts_input_ops = {
	.read	= ts_input_read,
	.fini	= ts_input_fini,
};

TSAPI struct tslib_module_info *mod_init(struct tsdev *dev, const char *params)
{
	struct tslib_input *i;

	i = malloc(sizeof(struct tslib_input));
	if (i == NULL)
		return NULL;

	i->module.ops = &__ts_input_ops;
	i->current_x = 0;
	i->current_y = 0;
	i->current_p = 0;
	i->sane_fd = 0;
	i->using_syn = 0;
	return &(i->module);
}
