/*
 *  tslib/plugins/linear.c
 *
 *  Copyright (C) 2001 Russell King.
 *
 * This file is placed under the LGPL.  Please see the file
 * COPYING for more details.
 *
 * $Id: linear.c,v 1.10 2005/02/26 01:47:23 kergoth Exp $
 *
 * Linearly scale touchscreen values
 */
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>

#include "tslib.h"
#include "tslib-filter.h"

//#define DEBUG
#define SIG_CALI	SIGRTMAX - 2

struct tslib_linear {
	struct tslib_module_info module;
	int	swap_xy;

// Linear scaling and offset parameters for pressure
	int	p_offset;
	int	p_mult;
	int	p_div;

// Linear scaling and offset parameters for x,y (can include rotation)
	int	a[7];
};

static int
linear_read(struct tslib_module_info *info, struct ts_sample *samp, int nr)
{
	struct tslib_linear *lin = (struct tslib_linear *)info;
	int ret, index;
	int xtemp,ytemp;

	ret = info->next->ops->read(info->next, samp, nr);
	if (ret >= 0) {
		int nr;

		for (nr = 0; nr < ret; nr++, samp++) {
			xtemp = samp->x; ytemp = samp->y;
#ifdef DEBUG
	fprintf(stdout, "RAW: x=%d, y=%d\n", xtemp, ytemp);
#endif
			samp->x = 	( lin->a[2] +
					lin->a[0]*xtemp + 
					lin->a[1]*ytemp ) >> 16;	//lin->a[6] == 65536
			samp->y =	( lin->a[5] +
					lin->a[3]*xtemp +
					lin->a[4]*ytemp ) >> 16; //lin->a[6] == 65536

			samp->pressure = ((samp->pressure + lin->p_offset)
						 * lin->p_mult) / lin->p_div;
			if (lin->swap_xy) {
				int tmp = samp->x;
				samp->x = samp->y;
				samp->y = tmp;
			}
#ifdef DEBUG
	fprintf(stdout, "CAL: x=%d, y=%d\n", samp->x, samp->y);
#endif	
		}
	}

	return ret;
}

static int linear_fini(struct tslib_module_info *info)
{
#ifdef DEBUG
	printf("******linear finit******\n");
#endif
	free(info);
	return 0;
}

static const struct tslib_ops linear_ops =
{
	.read	= linear_read,
	.fini	= linear_fini,
};

static int linear_xyswap(struct tslib_module_info *inf, char *str, void *data)
{
	struct tslib_linear *lin = (struct tslib_linear *)inf;

	lin->swap_xy = 1;
	return 0;
}

static const struct tslib_vars linear_vars[] =
{
	{ "xyswap",	(void *)1, linear_xyswap }
};

struct tslib_linear *lin;

static void check_cali(char *filename)
{
	struct stat sbuf;
	int pcal_fd;
	char pcalbuf[200];
	int index;
	char *tokptr;
	char *calfile=NULL;
	char *defaultcalfile = "/etc/pointercal";
	
#ifdef DEBUG
	printf("check calibration file: %s\n", filename);
#endif	
	/*
	 * Check calibration file
	 */
		
	calfile = filename;
	if(calfile == NULL || stat(calfile,&sbuf))
		calfile = defaultcalfile;
		
	if(!stat(calfile,&sbuf))
	{
		pcal_fd = open(calfile,O_RDONLY);
		read(pcal_fd,pcalbuf,200);
		lin->a[0] = atoi(strtok(pcalbuf," "));
		index=1;
		while(index<7) {
			tokptr = strtok(NULL," ");
			if(*tokptr!='\0') {
				lin->a[index] = atoi(tokptr);
				index++;
			}
		}
#ifdef DEBUG
		printf("Linear calibration constants: ");
		for(index=0;index<7;index++) printf("%d ",lin->a[index]);
		printf("\n");
#endif /*DEBUG*/
		close(pcal_fd);
	}
	
}


static char* get_name_from_pid(int pid)
{
	char cmd_f[128];
	char *buf = (char*)malloc(128);
	int  fd;
	int  filesize;

	sprintf(cmd_f, "/proc/%d/cmdline", pid);
	fd = open(cmd_f, O_RDONLY);
	if(fd < 0)
	{
		printf("open cmdline failed.\n");
		return NULL;
	}
	
	read(fd, buf, 128);
	close(fd);	
	return buf;
}

static int save_pid()
{
	pid_t pid;
	int   fd;
	char  *pname;
	char  buf[128] = {0};
	int   rc = 0;	

	pid = getpid();
	pname = get_name_from_pid(pid);
#ifdef DEBUG
	printf("current process : %s --> %d.\n", pname, pid);
#endif /*DEBUG*/
	
	if(pname && strstr(pname, "ts_calibrate") == NULL)
	{
		sprintf(buf, "%d", pid);
		fd = open("/etc/recalpid", O_CREAT | O_RDWR | O_TRUNC);
		write(fd, buf, strlen(buf) + 1);
		close(fd);
		chmod("/etc/recalpid", S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IROTH);
		rc = 1;
	}
	
	if(pname)
		free(pname);
#ifdef DEBUG
	printf("the ts_calibrate process, don't install the signal.\n");
#endif /*DEBUG*/
	return rc;

}

void new_op(int signum, siginfo_t *info, void *myact)
{
	int param = (int)(info->si_value.sival_int);
#ifdef DEBUG
	printf("recv the signal, the sig value is %d, parameter is %x\n", signum, param);
#endif /*DEBUG*/

	if(signum == SIG_CALI)
	{
		switch(param)
		{
			case 0x4C53: //LS-->landscape
				check_cali("/etc/pointercal_l");
				//setenv("TSLIB_CALIBFILE", "/etc/pointercal_l", 1);
				break;
			
			case 0x5052: //PR-->portrait
			default:
				check_cali("/etc/pointercal");
				//setenv("TSLIB_CALIBFILE", "/etc/pointercal", 1);
				break;
		}
	}
	else
	{
		printf("the signal value is not match.\n");
	}
}

#define NR_VARS (sizeof(linear_vars) / sizeof(linear_vars[0]))

TSAPI struct tslib_module_info *mod_init(struct tsdev *dev, const char *params)
{
	struct sigaction act;
	union  sigval mysigval;
	char *cal_file = getenv("TSLIB_CALIBFILE");
	
#ifdef DEBUG	
	printf("######linear init######\n");
#endif

	lin = malloc(sizeof(struct tslib_linear));
	if (lin == NULL)
		return NULL;

	lin->module.ops = &linear_ops;

// Use default values that leave ts numbers unchanged after transform
	lin->a[0] = -4115;
	lin->a[1] = 0;
	lin->a[2] = 50370000;
	lin->a[3] = 0;
	lin->a[4] = 4115;
	lin->a[5] = -414520;
	lin->a[6] = 65536;
	lin->p_offset = 0;
	lin->p_mult   = 1;
	lin->p_div    = 1;
	lin->swap_xy  = 0;

	
	check_cali(cal_file);	
		
	/*
	 * Parse the parameters.
	 */
	if (tslib_parse_vars(&lin->module, linear_vars, NR_VARS, params)) {
		free(lin);
		return NULL;
	}
	
	//save current procees's pid
	if(save_pid())
	{
		//install signal
		sigemptyset(&act.sa_mask);
		act.sa_sigaction = new_op;	
		act.sa_flags = SA_SIGINFO;
		if (sigaction(SIG_CALI, &act, NULL) < 0)
		{
#ifdef DEBUG
			printf("install signal error.\n");
#endif /*DEBUG*/
		}
		else
		{
#ifdef DEBUG
			printf("install signal success.\n");
#endif /*DEBUG*/
		}	
	}

	return &lin->module;
}
