/*
 *  tslib/tests/ts_calibrate.c
 *
 *  Copyright (C) 2001 Russell King.
 *
 * This file is placed under the GPL.  Please see the file
 * COPYING for more details.
 *
 * $Id: ts_calibrate.c,v 1.8 2004/10/19 22:01:27 dlowder Exp $
 *
 * Basic test program for touchscreen library.
 */
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <linux/fb.h>
#include <linux/input.h>
#include <errno.h>
#include "tslib.h"

#include "fbutils.h"
#include "testutils.h"

//#define USE_INPUT_API
#define SIG_CALI	SIGRTMAX - 2
#define min(a, b) (a < b ? a : b)

static int palette [] =
{
	0x000000, 0xffe080, 0xffffff, 0xe0c0a0
};

#define NR_COLORS (sizeof (palette) / sizeof (palette [0]))
#define CAL_NUM		9 	//5
#define DELTA1		20	//50
#define DELTA2		100

typedef struct {
	int x[CAL_NUM], xfb[CAL_NUM];
	int y[CAL_NUM], yfb[CAL_NUM];
	int a[7];
} calibration;

typedef struct __point{
	int x;
	int y;
}point;

point base_point_p[CAL_NUM] = {
	{11890,   416},
	{  228,   477},
	{  281, 16177},
	{11853, 16170},
	{10600,  1772},
	{ 1629,  1777},
	{ 1600, 14898},
	{10632, 14881},
	{ 6104,  8358}
};


point base_point_l[CAL_NUM] = {
	{  410,   521},
	{  506, 16207},
	{12088, 16204},
	{12043,   511},
	{ 1744,  1768},
	{ 1796, 14895},
	{10841, 14864},
	{10698,  1780},
	{ 6265,  8340}
};

#define DIFF	16 * 20

struct tsdev *ts = 0;
int last_x = -1, last_y = -1;

typedef void (*draw_cross_fun) (int x, int y, unsigned colidx) ;
typedef void (*draw_string_center_fun) (int x, int y, char *s, unsigned colidx);

draw_cross_fun  draw_cross = NULL;
draw_string_center_fun draw_string_center = NULL;

static int lock_ts(struct tsdev* ts_device, int is_lock)
{
	int ts_tmpfd = ts_fd(ts_device);
	if(ts_tmpfd == -1)
	{
		perror("ts_open");
		return 0;
	}
	
	unsigned long val = is_lock;
	int ioctl_ret = ioctl(ts_tmpfd, EVIOCGRAB, &val);
	if(ioctl_ret != 0)
	{
		printf("Error:%s\n", strerror(errno));
		return 0;
	}

	printf("lock the ts success\n");

	return 1;
}

static void sig(int sig)
{
	close_framebuffer ();
	fflush (stderr);
	printf ("signal %d caught\n", sig);
	fflush (stdout);
	exit (1);
}

int perform_calibration(calibration *cal) {
	int j;
	float n, x, y, x2, y2, xy, z, zx, zy;
	float det, a, b, c, e, f, i;
	float scaling = 65536.0;

// Get sums for matrix
	n = x = y = x2 = y2 = xy = 0;
	for(j=0;j<CAL_NUM;j++) {
		n += 1.0;
		x += (float)cal->x[j];
		y += (float)cal->y[j];
		x2 += (float)(cal->x[j]*cal->x[j]);
		y2 += (float)(cal->y[j]*cal->y[j]);
		xy += (float)(cal->x[j]*cal->y[j]);
	}

// Get determinant of matrix -- check if determinant is too small
	det = n*(x2*y2 - xy*xy) + x*(xy*y - x*y2) + y*(x*xy - y*x2);
	if(det < 0.1 && det > -0.1) {
		printf("ts_calibrate: determinant is too small -- %f\n",det);
		return 0;
	}

// Get elements of inverse matrix
	a = (x2*y2 - xy*xy)/det;
	b = (xy*y - x*y2)/det;
	c = (x*xy - y*x2)/det;
	e = (n*y2 - y*y)/det;
	f = (x*y - n*xy)/det;
	i = (n*x2 - x*x)/det;

// Get sums for x calibration
	z = zx = zy = 0;
	for(j=0;j<CAL_NUM;j++) {
		z += (float)cal->xfb[j];
		zx += (float)(cal->xfb[j]*cal->x[j]);
		zy += (float)(cal->xfb[j]*cal->y[j]);
	}

// Now multiply out to get the calibration for framebuffer x coord
	cal->a[0] = (int)((a*z + b*zx + c*zy)*(scaling));
	cal->a[1] = (int)((b*z + e*zx + f*zy)*(scaling));
	cal->a[2] = (int)((c*z + f*zx + i*zy)*(scaling));

	printf("%f %f %f\n",(a*z + b*zx + c*zy),
				(b*z + e*zx + f*zy),
				(c*z + f*zx + i*zy));

// Get sums for y calibration
	z = zx = zy = 0;
	for(j=0;j<CAL_NUM;j++) {
		z += (float)cal->yfb[j];
		zx += (float)(cal->yfb[j]*cal->x[j]);
		zy += (float)(cal->yfb[j]*cal->y[j]);
	}

// Now multiply out to get the calibration for framebuffer y coord
	cal->a[3] = (int)((a*z + b*zx + c*zy)*(scaling));
	cal->a[4] = (int)((b*z + e*zx + f*zy)*(scaling));
	cal->a[5] = (int)((c*z + f*zx + i*zy)*(scaling));

	printf("%f %f %f\n",(a*z + b*zx + c*zy),
				(b*z + e*zx + f*zy),
				(c*z + f*zx + i*zy));

// If we got here, we're OK, so assign scaling to a[6] and return
	cal->a[6] = (int)scaling;
	return 1;
/*	
// This code was here originally to just insert default values
	for(j=0;j<7;j++) {
		c->a[j]=0;
	}
	c->a[1] = c->a[5] = c->a[6] = 1;
	return 1;
*/

}

static int save_coordinate(char * name, int x, int y)
{
	//char buf[128];
	FILE* fp;
	
	//sprintf(buf, "%s:  %4d  %4d\n", name, cal->x [index], cal->y [index]);
	
	fp = fopen("/etc/coord", "a+");
	if(fp)
	{
		fprintf(fp, "%s: x=%05d y=%05d\r\n", name, x, y);
		fclose(fp);
	}
	return 0;

}

//#define GET_RAW_COORDINATE
#ifdef GET_RAW_COORDINATE
static void get_sample (struct tsdev *ts, calibration *cal,
			int index, int x, int y, char *name, int blandscape)
{
	static int last_x = -1, last_y;

	if (last_x != -1) {
#define NR_STEPS 10
		int dx = ((x - last_x) << 16) / NR_STEPS;
		int dy = ((y - last_y) << 16) / NR_STEPS;
		int i;
		last_x <<= 16;
		last_y <<= 16;
		for (i = 0; i < NR_STEPS; i++) {
			draw_cross (last_x >> 16, last_y >> 16, 2 | XORMODE);
			usleep (2000);
			draw_cross (last_x >> 16, last_y >> 16, 2 | XORMODE);
			last_x += dx;
			last_y += dy;
		}
	}

	draw_cross(x, y, 2 | XORMODE);
	getxy (ts, &cal->x [index], &cal->y [index]);
	draw_cross(x, y, 2 | XORMODE);
	
	last_x = x;
	last_y = y;
	
	if(blandscape)
	{
		cal->xfb [index] = xres - y;
		cal->yfb [index] = x;
	}
	else
	{
		cal->xfb [index] = x;
		cal->yfb [index] = y;
	}

	printf("%s : X = %4d Y = %4d\n", name, cal->x [index], cal->y [index]);
	save_coordinate(name, cal->x [index], cal->y [index]);
}

#else

static void get_sample (struct tsdev *ts, calibration *cal,
			int index, int x, int y,  char *name, int blandscape)
{
	int last_x_tmp = last_x, last_y_tmp = last_y;
	point * base_point = NULL;
	
	if(blandscape)
		base_point = base_point_l;
	else
		base_point = base_point_p;
	
	do{
		if (last_x != -1) {
#define NR_STEPS 10
			int dx = ((x - last_x) << 16) / NR_STEPS;
			int dy = ((y - last_y) << 16) / NR_STEPS;
			int i;
			last_x <<= 16;
			last_y <<= 16;
			for (i = 0; i < NR_STEPS; i++) {
				draw_cross (last_x >> 16, last_y >> 16, 2 | XORMODE);
				usleep (1000 * 3);
				draw_cross (last_x >> 16, last_y >> 16, 2 | XORMODE);
				last_x += dx;
				last_y += dy;
			}
		}
		
		draw_cross(x, y, 2 | XORMODE);
		getxy (ts, &cal->x [index], &cal->y [index]);
		draw_cross(x, y, 2 | XORMODE);
		
		
		if(last_x_tmp == -1)
		{
			last_x = 0;
			last_y = 0;
		}
		else
		{
			last_x = last_x_tmp;
			last_y = last_y_tmp;
		}
		
	}while( cal->x [index] >= base_point[index].x + DIFF ||
			cal->x [index] <= base_point[index].x - DIFF ||
			cal->y [index] >= base_point[index].y + DIFF ||
			cal->y [index] <= base_point[index].y - DIFF   );
	
	last_x = x;
	last_y = y;
	
	if(blandscape)
	{
		cal->xfb [index] = xres - y;
		cal->yfb [index] = x;
	}
	else
	{
		cal->xfb [index] = x;
		cal->yfb [index] = y;
	}
	
	printf("%s : X = %d Y = %d\n", name, cal->x [index], cal->y [index]);
}

#endif

static int read_pid()
{
	int   pid;
	int   fd;
	int   filesize;
	char  *buf;
	char  str[128];	
	//char  *ptr1, *ptr2;
	struct stat sbuf;

	pid = (int)getpid();
	printf("ts_calibreate: pid is %d.\n", pid);

	fd = open("/etc/recalpid", O_RDONLY);
	if(fd < 0)
	{
		printf("open file failed.\n");
		return fd;
	}

	filesize = lseek(fd, 0, SEEK_END);
	buf = (char*) malloc(filesize + 1);
	lseek(fd, 0, SEEK_SET);
	read(fd, buf, filesize);
	close(fd);

	pid = atoi(buf);
	printf("read pid from pidfile is %d.\n", pid);
	free(buf);
	
	sprintf(str, "/proc/%d/", pid);	

	if(stat(str, &sbuf) != 0)
	{
		printf("the process %d isn't exsit.\n", pid);
		return 0;
	}
	
	return pid;
} 


static int clearbuf(struct tsdev *ts)
{
	int fd = ts_fd(ts);
	fd_set fdset;
	struct timeval tv;
	int nfds;
	struct ts_sample sample;

	while (1) {
		FD_ZERO(&fdset);
		FD_SET(fd, &fdset);

		tv.tv_sec = 0;
		tv.tv_usec = 0;

		nfds = select(fd + 1, &fdset, NULL, NULL, &tv);
		if (nfds == 0) break;

		if (ts_read_raw(ts, &sample, 1) < 0) {
			perror("ts_read");
			exit(1);
		}
	}
}

int main(int argc, char **argv)
{
	struct stat sbuf;
	unsigned int i;
	int pid;
	sigval_t mysigval;
	calibration cal;
	int cal_fd;
	char cal_buffer[256];
	char *tsdevice = NULL;
	char *calfile = NULL;
	char *calfile_p = "/etc/pointercal";
	char *calfile_l = "/etc/pointercal_l";
	int bexsitcalfile = 0;
	int blandscape = 0;
	int xres_tmp = 0, yres_tmp = 0;
	
	signal(SIGSEGV, sig);
	signal(SIGINT, sig);
	signal(SIGTERM, sig);
	
	if(argc > 1 && (!strcasecmp(argv[1], "-l") || !strcasecmp(argv[1], "--l")))
	{
		blandscape = 1;
	}

	if( (tsdevice = getenv("TSLIB_TSDEVICE")) != NULL ) {
		ts = ts_open(tsdevice,0);
	} else {
#ifdef USE_INPUT_API
		ts = ts_open("/dev/input/event0", 0);
#else
		ts = ts_open("/dev/touchscreen/ucb1x00", 0);
#endif /* USE_INPUT_API */
	}

	if (!ts) {
		perror("ts_open");
		exit(1);
	}

	//lock_ts(ts, 1);

	if (ts_config(ts)) {
		perror("ts_config");
		//lock_ts(ts, 0);
		exit(1);
	}

	if (open_framebuffer()) {
		close_framebuffer();
		//lock_ts(ts, 0);
		exit(1);
	}
	
	if(blandscape) //landscape
	{	
		xres_tmp = yres;
		yres_tmp = xres;
		draw_cross = put_cross_l;
		draw_string_center = put_string_center_l;
		calfile = calfile_l;
		mysigval.sival_int = 0x4C53; //LS-->landscape
		
	}
	else //portrait
	{
		xres_tmp = xres;
		yres_tmp = yres;
		draw_cross = put_cross;
		draw_string_center = put_string_center;
		calfile = calfile_p;
		mysigval.sival_int = 0x5052; //PR-->portrait
	}
				

	for (i = 0; i < NR_COLORS; i++)
		setcolor (i, palette [i]);
	
	draw_string_center (xres_tmp / 2, yres_tmp / 4,
			   "TSLIB calibration utility", 2);
	draw_string_center (xres_tmp / 2, yres_tmp / 4 + 20,
			   "Touch crosshair to calibrate", 2);
	
	printf("xres = %d, yres = %d\n", xres_tmp, yres_tmp);

// Read a touchscreen event to clear the buffer
	//getxy(ts, 0, 0);

	 // Clear the buffer
	clearbuf(ts);
	get_sample (ts, &cal, 0, DELTA1,  DELTA1, "Top left", blandscape);
	
	clearbuf(ts);
	get_sample (ts, &cal, 1, xres_tmp - DELTA1, DELTA1, "Top right", blandscape);
	
	clearbuf(ts);
	get_sample (ts, &cal, 2, xres_tmp - DELTA1, yres_tmp - DELTA1, "Bot right", blandscape);
	
	clearbuf(ts);
	get_sample (ts, &cal, 3, DELTA1, yres_tmp - DELTA1, "Bot left", blandscape);
	
	if(CAL_NUM > 5)
	{
		clearbuf(ts);
		get_sample (ts, &cal, 4, DELTA2, DELTA2, "Top left2", blandscape);
		
		clearbuf(ts);
		get_sample (ts, &cal, 5, xres_tmp - DELTA2, DELTA2, "Top right2", blandscape);
		
		clearbuf(ts);
		get_sample (ts, &cal, 6, xres_tmp - DELTA2, yres_tmp - DELTA2, "Bot right2", blandscape);
		
		clearbuf(ts);
		get_sample (ts, &cal, 7, DELTA2, yres_tmp - DELTA2, "Bot left2", blandscape);
	}
	
	clearbuf(ts);
	get_sample (ts, &cal, 8, xres_tmp / 2, yres_tmp / 2, "Center", blandscape);

	if (perform_calibration (&cal)) {
		printf ("Calibration constants: ");
		for (i = 0; i < 7; i++) printf("%d ", cal.a [i]);
		printf("\n");
		//if ((calfile = getenv("TSLIB_CALIBFILE")) == NULL)
		//	calfile = defaultcalfile;
		
		if(stat(calfile,&sbuf)==0)
		{
			printf("the calibration file is exist\n");
			bexsitcalfile = 1;
		}

		cal_fd = open (calfile, O_CREAT | O_RDWR);

		sprintf (cal_buffer,"%d %d %d %d %d %d %d",
			 cal.a[1], cal.a[2], cal.a[0],
			 cal.a[4], cal.a[5], cal.a[3], cal.a[6]);
		write (cal_fd, cal_buffer, strlen (cal_buffer) + 1);
		close (cal_fd);
                i = 0;
	} else {
		printf("Calibration failed.\n");
		i = -1;
	}

	if (bexsitcalfile && !blandscape)
	{
		if ((pid = read_pid()) > 0)
		{
			if (sigqueue(pid, SIG_CALI, mysigval) == -1)
				printf("send signal error\n");
			else
				printf("send signal success\n");
		}
		
	}

	close_framebuffer();
	//lock_ts(ts, 0);
	return i;
}
